#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <stdlib.h>
#include <ctype.h>

#define  IKORG_PCG_SIZE  164
#define  IKORG_DRUMKIT_SIZE     0xa50

#define  KORGi3_LASTMULTISOUND  340
#define  KORGi3_LASTDRUMSOUND   163

typedef struct
{
  long adr, len;
} block;

#define BLOCK_NAMECOMBI    0   // A+B
#define BLOCK_NAMEPROG    1   // A+B
#define BLOCK_GLOBAL   2
#define BLOCK_DRUMKIT1 3
#define BLOCK_COMBI1   4  // 0-99
#define BLOCK_PROG1   5  // 200-299
#define BLOCK_DRUMKIT2 6
#define BLOCK_COMBI2   7  // 100-199  0x88
#define BLOCK_PROG2   8  // 300-399  0xA4

typedef struct
{
  char korg[4]; // KORG
  char version;  // '5' ==> N-series '9' => i-series
  char zero[11];
  block b[9];
} N_PCGHEADER;

N_PCGHEADER h;

#define MAXPROG 500
char progname[2][500][10];
int progsize[2];
int datacnt[9];

char newname[128] = "split?.pcg";

void copybytes(FILE* f, FILE* t, long len)
{
   while (len-- > 0)
     fputc(fgetc(f), t);
}

typedef struct
{
  char name[10];
  char doubleoscillator;
  char monophonic : 1;
  char hold : 1;
  int  osc1_multisound;
  char row1[1];
  int  osc2_multisound;

  char not_important[147];
} PROGDATA;

typedef struct
{
  int drumsound;
  unsigned char key;
  unsigned char panlevel : 5;
  unsigned char group : 3;
  char d1[2];
  signed char level;
  char d2;
  signed char decay;
  unsigned char effect_sendlevel; // C-D
  char d3[12];
} PCGDRUM;

static char* basekey = "CCDDEFFGGAB";

const char* note2str(int note)
{
int oct = note / 12 - 1;
static char* nstr = "C#-4";
char* p = nstr;
static char* basekey = "CCDDEFFGGAAB";

  note %= 12;

  *p++ = basekey[note];
  switch(note)
  {
  case 1: //c#
  case 3: //d#
  case 6: //f#
  case 8: //g#
  case 10: //a#
     *p++ = '#';
     break;
  }
  sprintf(p, "%d", oct);
  return (const char*)nstr;
}

void copydrumkit(FILE* f, FILE* t, long len)
{
  if (len != IKORG_DRUMKIT_SIZE)
    copybytes(f, t, len);
  else
  {
    assert(sizeof(PCGDRUM) == 22);
    PCGDRUM drum;

    while (len > 0)
    {
      if (fread(&drum, sizeof(drum), 1, f) != 1)
      {
        fprintf(stderr, "unexpected end of file, input file is too short.\n");
        exit(1);
      }
      if (drum.drumsound > 163)
      {
        printf("Warning: drumkit uses drumsound %d at key %s\n", drum.drumsound, note2str(drum.key));
      }
      if (fwrite(&drum, sizeof(drum), 1, t) != 1)
      {
        fprintf(stderr, "write error, maybe disk full\n");
        exit(1);
      }
      len -= sizeof(PCGDRUM);
    }
  }
}

void copyblock(FILE* f, FILE* t, block& fblock, block& tblock)
{
   fseek(f, fblock.adr, SEEK_SET);
   tblock.len = fblock.len;
   tblock.adr = ftell(t);

   if (fblock.len == IKORG_DRUMKIT_SIZE)
     copydrumkit(f, t, fblock.len);
   else
     copybytes(f, t, fblock.len);
}

void copyprogram(FILE* f, FILE* t, int len, long progpos)
{
  if (len != IKORG_PCG_SIZE)
    copybytes(f, t, len);
  else
  {
  PROGDATA prog;

    assert(sizeof(prog) == IKORG_PCG_SIZE);

    if (fread(&prog, sizeof(prog), 1, f) != 1)
    {
      fprintf(stderr, "unexpected end of file found\n");
      exit(1);
    }

    if (prog.osc1_multisound > KORGi3_LASTMULTISOUND ||
        (prog.doubleoscillator && prog.osc2_multisound > KORGi3_LASTMULTISOUND))
    {
      printf("Warning: %10.10s uses multisounds %s  %3d %3d",
       prog.name, prog.doubleoscillator ? "double" : "single",
       prog.osc1_multisound, prog.osc2_multisound);

       if (prog.name[0] != '?')
       {
         memmove(prog.name+1, prog.name, sizeof(prog.name)-1);
         prog.name[0] = '?';
         printf("  -> %10.10s", prog.name);
       }
       printf("\n");

       long oldpos = ftell(t);
       fseek(t, progpos, SEEK_SET);
       fwrite(prog.name, sizeof(prog.name), 1, t);
       fseek(t, oldpos, SEEK_SET);
    }
    if (fwrite(&prog, sizeof(prog), 1, t) != 1)
    {
      fprintf(stderr, "write error, maybe disk full\n");
      exit(1);
    }
  }
}

void splitpcg(FILE* f, int nr, int blockname, int nameidx, int blockprog, int progidx, int programcnt)
{
N_PCGHEADER newh = h;
int mode = blockname - BLOCK_NAMECOMBI; // mode 0: combi, mode 1: programs
int group = (blockprog == BLOCK_COMBI1 || blockprog == BLOCK_PROG1) ? 1 : 2;

  if (nr > 10)
   return;
  char* p = strchr(newname, '.')-1;
  *p = nr + '0';

  printf("%s mode %d prog %d count %d\n", newname, mode, nameidx, programcnt);

  FILE* t = fopen(newname, "wb");
  if (!t)
  {
    perror(newname);
    return;
  }
  fwrite(&newh, sizeof(newh), 1, t);
  newh.b[BLOCK_NAMECOMBI].adr = ftell(t);
  newh.b[BLOCK_NAMECOMBI].len = 0;
  newh.b[BLOCK_NAMEPROG].adr = ftell(t); // i3 names  64+2
  newh.b[BLOCK_NAMEPROG].len = 66 * 10;

  int i;
  int cnt = programcnt;
  for (i = nameidx; cnt > 0; cnt--)
  {
    fwrite(progname[mode][i], 10, 1, t);
    i++;
  }

  for (i = programcnt; i < 66; i++)
    fwrite("Empty     ", 10, 1, t);

  copyblock(f, t, h.b[BLOCK_GLOBAL], newh.b[BLOCK_GLOBAL]);
  copyblock(f, t, h.b[group == 2 ? BLOCK_DRUMKIT2 : BLOCK_DRUMKIT1], newh.b[BLOCK_DRUMKIT1]);

  newh.b[BLOCK_COMBI1].adr = ftell(t);
  newh.b[BLOCK_COMBI1].len = 0;
  newh.b[BLOCK_PROG1].adr = ftell(t);
  newh.b[BLOCK_PROG1].len = 66L * (long)progsize[mode];

  cnt = programcnt;
  for (i = progidx; cnt > 0; cnt--)
  {
    fseek(f, h.b[blockprog].adr + i * progsize[mode], SEEK_SET);
    copyprogram(f, t, progsize[mode], newh.b[BLOCK_NAMEPROG].adr + (i-progidx)*10);
    i++;
  }
  for (i  =programcnt; i <  66; i++)
  {
    fwrite("Empty     ", 10, 1, t);
    int len = progsize[mode] - 10;
    while (len-- > 0)
      fputc(0, t);

  }
  newh.b[BLOCK_DRUMKIT2].adr = ftell(t);
  newh.b[BLOCK_DRUMKIT2].len = 0;
  newh.b[BLOCK_COMBI2].adr = ftell(t);
  newh.b[BLOCK_COMBI2].len = 0;
  newh.b[BLOCK_PROG2].adr = ftell(t);
  newh.b[BLOCK_PROG2].len = 0;

  rewind(t);
  fwrite(&newh, sizeof(newh), 1, t);
  fclose(t);
}

void splitlist(FILE* f, int blockname, int blockprog)
{
static int splitnr = 1;

  int mode = blockname-BLOCK_NAMECOMBI;
  if (progsize[mode] != IKORG_PCG_SIZE)
    return;

  int count = progsize[mode] == 0 ? 0 : h.b[blockprog].len / progsize[mode];

  int i = 0;
  int namei = 0;
  if (blockprog == BLOCK_COMBI2)
    namei = datacnt[BLOCK_COMBI1];
  else if (blockprog == BLOCK_PROG2)
    namei = datacnt[BLOCK_PROG2];
  while (count > 0)
  {
    int n = 64;
    if (count < n)
      n = count;
    splitpcg(f, splitnr++, blockname, namei, blockprog, i, n);
    i+= n;
    namei += n;
    count -=n;
  }
}

char* skippath(char* p)
{
  if (!p)
    return 0;
  char* p1 = strrchr(p, '\\');
  if (p1)
    p = p1+1;
  p1 = strrchr(p, '/');
  if (p1)
    p = p1+1;
  p1 = strrchr(p, ':');
  if (p1)
    p = p1+1;
  return p;
}

void main(int argc, char** argv)
{
  argc--; argv++;

  if (argc == 0)
  {
    fprintf(stderr, "usage: splitpcg filename.pcg\n");
    return;
  }
  FILE* f = fopen(*argv, "rb");
  if (!f)
  {
    perror(*argv);
    return;
  }

  fread(&h, sizeof(h), 1, f);
  if (strncmp(h.korg, "KORG", 4) != 0)
  {
    fprintf(stderr, "%s is not a korg pcg file\n", *argv);
    return;
  }

  char* p = skippath(*argv);
  if (!p)
     return;
  strcpy(newname, p);
  p = strchr(newname, '.');
  if (p)
    *p = 0;
  int splitnamelen = strlen(newname);
  if (splitnamelen < 8)
    strcpy(p, "?.pcg");
  else
  {
    if (!isdigit(p[-1]))
    {
      strcpy(p-1, "?.pcg");
    }
    else
      strcpy(newname, "split?.pcg");
  }

  datacnt[BLOCK_NAMECOMBI] = (int)(h.b[BLOCK_NAMECOMBI].len / 10);
  datacnt[BLOCK_NAMEPROG] = (int)(h.b[BLOCK_NAMEPROG].len / 10);

  if (datacnt[BLOCK_NAMECOMBI] > MAXPROG)
     datacnt[BLOCK_NAMECOMBI] = MAXPROG;
  if (datacnt[BLOCK_NAMEPROG] > MAXPROG)
     datacnt[BLOCK_NAMEPROG] = MAXPROG;

  fseek(f, h.b[BLOCK_NAMECOMBI].adr, SEEK_SET);
  fread(progname[0], 10, datacnt[BLOCK_NAMECOMBI], f);
  fseek(f, h.b[BLOCK_NAMEPROG].adr, SEEK_SET);
  fread(progname[1], 10, datacnt[BLOCK_NAMEPROG], f);

  if (datacnt[BLOCK_NAMECOMBI] == 0)
    progsize[0] = 0;
  else
    progsize[0] = (int)((h.b[BLOCK_COMBI1].len+h.b[BLOCK_COMBI2].len) / datacnt[BLOCK_NAMECOMBI]);
  if (datacnt[BLOCK_NAMEPROG] == 0)
    progsize[1] = 0;
  else
    progsize[1] = (int)((h.b[BLOCK_PROG1].len+h.b[BLOCK_PROG2].len) / datacnt[BLOCK_NAMEPROG]);


  datacnt[BLOCK_COMBI1] = progsize[0] == 0 ? 0 : h.b[BLOCK_COMBI1].len / progsize[0];
  datacnt[BLOCK_COMBI2] = progsize[0] == 0 ? 0 : h.b[BLOCK_COMBI2].len / progsize[0];
  datacnt[BLOCK_PROG1] = progsize[1] == 0 ? 0 : h.b[BLOCK_PROG1].len / progsize[1];
  datacnt[BLOCK_PROG2] = progsize[1] == 0 ? 0 : h.b[BLOCK_PROG2].len / progsize[1];

  splitlist(f, BLOCK_NAMECOMBI, BLOCK_COMBI1);
  splitlist(f, BLOCK_NAMECOMBI, BLOCK_COMBI2);
  splitlist(f, BLOCK_NAMEPROG, BLOCK_PROG1);
  splitlist(f, BLOCK_NAMEPROG, BLOCK_PROG2);
  fclose(f);
}